
/*
 * Field Magnetometer
 * GPS, compass, OLED, all 7 Mag-Sensors, SD, gps calculation for sensors on SD card
 * =====================================================================================
 * 
 * Markus Opitz 2022
 * https://www.instructables.com/member/Markus%20Opitz/
 * 
 *
 * SD card reader:
 * 5v – 5v
 * cs –   53 (yellow)
 * mosi – 51 (blue)
 * sck –  52 (green)
 * miso – 50 (violet)
 * gnd – gnd
*/

  
//GPS
//Connect with pin 18 and 19
#include <TinyGPS.h>
//long   lat,lon; // create variable for latitude and longitude object
float lat,lon;
TinyGPS gps; // create gps object
float lat1, lat2, lat3, lat4, lat5, lat6, lat7;
float lon1, lon2, lon3, lon4, lon5, lon6, lon7;
float distance = 0.25;  //distance between sensors = 0.25m
float x1, x2, x3, y1, y2, y3;

// Compass
#include <Wire.h>
#include <LSM303.h>
LSM303 compass;
float heading;

//Mag-Sensors
const float hall1 = A1, hall2 = A2, hall3 = A3, hall4 = A4, hall5 = A5, hall6 = A6, hall7 = A7;
float sum1 = 0, sum2 = 0, sum3 = 0, sum4 = 0, sum5 = 0, sum6 = 0, sum7 = 0;
float calHall1 = 0, calHall2 = 0, calHall3 = 0, calHall4 = 0, calHall5 = 0, calHall6 = 0, calHall7 = 0;


//OLED
#include <SPI.h>
#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SH1106.h>
#define OLED_RESET 4
Adafruit_SH1106 display(OLED_RESET);
#define NUMFLAKES 10
#define XPOS 0
#define YPOS 1
#define DELTAY 2
#if (SH1106_LCDHEIGHT != 64)
#error("Height incorrect, please fix Adafruit_SH1106.h!");
#endif

//SD
#include <SPI.h>
#include <SD.h>
File myFile;
int chipselect  = 10;
String fileName = "magData.csv";


void setup(){   // *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  
Serial.begin(57600); // connect serial
Serial.println("STARTING MAGNETOMETER"); Serial.println("");
Serial.println("latitude,longitude,magnet,x,y");

//OLED
  display.begin(SH1106_SWITCHCAPVCC, 0x3C);  // initialize with the I2C addr 0x3D (for the 128x64)
  display.clearDisplay();
  display.display();
  delay(1000);
  // Clear the buffer.
  display.clearDisplay();
  display.display();
  delay(1000);
  Serial.println("OLED buffer clear");
  OLEDstart();
   display.setCursor(6, 50);
   display.setTextSize(1);
   display.println("Explore the ground!");
   display.display();
  delay(2000);

  display.clearDisplay();
  //display.display();
  OLEDstart();
  OLEDwait();
  display.display();
  
// Compass
  Wire.begin();
  compass.init();
  compass.enableDefault();
  delay(10);
  Serial.println("compass init");
  /*
   use calibrate_compass.ino for the exact adjustment of the compass
  */
  
  compass.m_min = (LSM303::vector<int16_t>){  -343,   -444,   -513};
  compass.m_max = (LSM303::vector<int16_t>){  +510,   +365,   +296};

//Mag-Sensors
  pinMode(hall1, INPUT);
  pinMode(hall2, INPUT);
  pinMode(hall3, INPUT);
  pinMode(hall4, INPUT);
  pinMode(hall5, INPUT);
  pinMode(hall6, INPUT);
  pinMode(hall7, INPUT);
  calHall(); // get value for calibration

//GPS
Serial1.begin(9600); // connect gps
Serial.println("GPS init");
delay(30);


//SD
  //if (!SD.begin(chipselect)) {
  if (!SD.begin()) {
    Serial.println("Micro SD reading failed!");
    Serial.println("* Is connection correct?");
    Serial.println("* Memory card is not inserted?");
    Serial.println("\n");
    
    //OLED
      display.clearDisplay();
      display.setCursor(4,15);
      display.setTextSize(1);
      display.println("Micro SD reading");
      display.setCursor(4,25);
      display.println("failed!");
      // no SD
      display.fillRect(3, 50, 8, 17, WHITE);
      display.fillRect(11, 55, 5, 12, WHITE);
      display.fillTriangle(11, 50, 15, 54, 12, 54 ,WHITE);
      display.setCursor(4, 54);
      display.setTextColor(BLACK);
      display.setTextSize(1);
      display.println("SD");
      display.drawLine(1, 47, 19, 63, BLACK);
      display.drawLine(1, 48, 19, 64, BLACK);
 
      display.display();

    while (1) {
      delay(10);
    }
  } 
  else {
    Serial.println("The MicroSD card has been read.");
  }

  if (SD.exists(fileName)) {
    SD.remove(fileName); // delete file if it already exists
  }
  
  Serial.println("Open file, writing headline");
  myFile = SD.open(fileName, FILE_WRITE);
  myFile.println("latitude,longitude,magnet,heading,x,y");
  myFile.close();
  Serial.println("\n");

}
 
void loop(){  // *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *

//  get GPS
   while(Serial1.available()){ // check for gps data
    //Serial.println("GPS active");
    if(gps.encode(Serial1.read()))// encode gps data
    { 
    gps.f_get_position(&lon,&lat); // get latitude and longitude

    //Serial.print("Longitude: ");
    Serial.print(lon,8);
    Serial.print(",");
    //Serial.print("Latitude: ");
    Serial.print(lat,8);

    Serial.print(",");

    getHall();  //Hall values
    comp();     //compass heading
    Serial.println("go calculating...");
    gpsCalculate(); //calculate hall sensor positions
    saveSD();       //save to SD
    showOLED();     //dispaly
    }
   }
} 



void comp() {
   compass.read();
   heading = compass.heading();
   Serial.println(heading,0);
   delay(10);
}


void calHall() {    //calibrate the sensors to one level at beginning
  // make an average out of 20 readings for each sensor to avoid inaccuracies
  for(int i = 0; i < 20; i++){      
      float value1 = analogRead(hall1);
      float value2 = analogRead(hall2);
      float value3 = analogRead(hall3);
      float value4 = analogRead(hall4);
      float value5 = analogRead(hall5);
      float value6 = analogRead(hall6);
      float value7 = analogRead(hall7);
      
      sum1 += value1;
      sum2 += value2;
      sum3 += value3;
      sum4 += value4;
      sum5 += value5;
      sum6 += value6;
      sum7 += value7;
  }
  sum1 = sum1 / 20;
  sum2 = sum2 / 20;
  sum3 = sum3 / 20;
  sum4 = sum4 / 20;
  sum5 = sum5 / 20;
  sum6 = sum6 / 20;
  sum7 = sum7 / 20;

// get calibration value
 calHall1 = (550 - sum1);
 calHall2 = (550 - sum2);
 calHall3 = (550 - sum3);
 calHall4 = (550 - sum4);
 calHall5 = (550 - sum5);
 calHall6 = (550 - sum6);
 calHall7 = (550 - sum7);
}

void getHall() {
  //Serial.println("get sensor value");
  // make an average out of 20 readings for each sensor to avoid inaccuracies
  for(int i = 0; i < 20; i++){      
      float value1 = analogRead(hall1);
      float value2 = analogRead(hall2);
      float value3 = analogRead(hall3);
      float value4 = analogRead(hall4);
      float value5 = analogRead(hall5);
      float value6 = analogRead(hall6);
      float value7 = analogRead(hall7);
      
      sum1 += value1;
      sum2 += value2;
      sum3 += value3;
      sum4 += value4;
      sum5 += value5;
      sum6 += value6;
      sum7 += value7;
  }
  sum1 = sum1 / 20;
  sum2 = sum2 / 20;
  sum3 = sum3 / 20;
  sum4 = sum4 / 20;
  sum5 = sum5 / 20;
  sum6 = sum6 / 20;
  sum7 = sum7 / 20;

// calibrate the sensors with the value from setup();
  sum1 = sum1 + calHall1 +50 ; //zur Markierung von Sensor 1
  sum2 = sum2 + calHall2;
  sum3 = sum3 + calHall3;
  sum4 = sum4 + calHall4;
  sum5 = sum5 + calHall5;
  sum6 = sum6 + calHall6;
  sum7 = sum7 + calHall7;
  

/*
  // print results
  Serial.print(sum1);Serial.print("     ");Serial.print(sum2); Serial.print("     ");Serial.print(sum3); Serial.print("         ");
  Serial.print(sum4); Serial.print("         ");
  Serial.print(sum5);Serial.print("     ");Serial.print(sum6); Serial.print("     ");Serial.println(sum7);
*/  
}


void showOLED() {
  //Serial.println("OLED active");
  display.clearDisplay();
  
  // GPS
  display.setCursor(4,3);
  display.setTextColor(BLACK);
  display.setTextSize(2);
  display.fillCircle(9, 9, 9, WHITE);
  display.println("G");

// Sensors
  display.fillRect(50, 2, 7, (sum1-300)*0.1, WHITE); 
  display.fillRect(60, 2, 7, (sum2-300)*0.1, WHITE);
  display.fillRect(70, 2, 7, (sum3-300)*0.1, WHITE);
  display.fillRect(80, 2, 7, (sum4-300)*0.1, WHITE);
  display.fillRect(90, 2, 7, (sum5-300)*0.1, WHITE);
  display.fillRect(100, 2, 7, (sum6-300)*0.1, WHITE);
  display.fillRect(110, 2, 7, (sum7-300)*0.1, WHITE);
  display.setTextSize(1); display.setTextColor(BLACK);
  display.setCursor(51,4); display.println("1");
  display.setCursor(61,4); display.println("2");
  display.setCursor(71,4); display.println("3");
  display.setCursor(81,4); display.println("4");  
  display.setCursor(91,4); display.println("5");
  display.setCursor(101,4); display.println("6");
  display.setCursor(111,4); display.println("7"); 
    

//SD-card
  display.fillRect(3, 50, 8, 17, WHITE);
  display.fillRect(11, 55, 5, 12, WHITE);
  display.fillTriangle(11, 50, 15, 54, 12, 54 ,WHITE);
  display.setCursor(4, 54);
  display.setTextColor(BLACK);
  display.setTextSize(1);
  display.println("SD");
  

// Heading
  display.setCursor(2,33);
  display.setTextColor(WHITE);
  display.setTextSize(1);
  display.println(heading,0);
  display.drawCircle(24, 34, 1, WHITE);

  display.display();  // display all on OLED 
}

void OLEDstart() {
  display.setTextColor(WHITE);
  display.setCursor(3, 2);
  display.setTextSize(2);
  display.println("Magneto");
  display.setCursor(65, 22);
  display.println("meter");
  display.fillRect(6, 25, 6, 10, WHITE);
  display.fillRect(15, 30, 6, 5, WHITE);
  display.fillRect(24, 27, 6, 8, WHITE);
  display.fillRect(33, 32, 6, 3, WHITE);
  display.fillRect(42, 29, 6, 6, WHITE);
}

void OLEDwait() {
  display.setCursor(6, 45);
  display.setTextSize(1);
  display.println("started,");
  display.setCursor(6, 55);
  display.println("waiting for GPS");
}

void saveSD(){
     myFile = SD.open(fileName, FILE_WRITE);
   Serial.println("Writing data...");

    myFile.print(lon1,9);
    myFile.print(",");
    myFile.print(lat1,9);
    myFile.print(","); 
    myFile.print(sum1,1);
     myFile.print(",");
     myFile.print(heading);
     myFile.print(",");
     myFile.print(x3,9);
     myFile.print(","); 
     myFile.println(y3,9);

    myFile.print(lon2,9);
    myFile.print(",");
    myFile.print(lat2,9);
    myFile.print(",");
    myFile.print(sum2,1);
     myFile.print(",");
     myFile.print(heading);
     myFile.print(","); 
     myFile.print(x2,9);
     myFile.print(","); 
     myFile.println(y2,9);    

    myFile.print(lon3,9);
    myFile.print(",");
    myFile.print(lat3,9);
    myFile.print(",");
    myFile.print(sum3,1);
     myFile.print(",");
     myFile.print(heading);
     myFile.print(","); 
     myFile.print(x1,9);
     myFile.print(","); 
     myFile.println(y1,9);    
   
    myFile.print(lon,9); // center sensor
    myFile.print(",");
    myFile.print(lat,9);
    myFile.print(",");
    myFile.print(sum4,1);
     myFile.print(",");
     myFile.print(heading);
     myFile.print(","); 
     myFile.print("0.0");
     myFile.print(","); 
     myFile.println("0.0");      

    myFile.print(lon5,9);
    myFile.print(",");
    myFile.print(lat5,9);
    myFile.print(",");
    myFile.print(sum5,1);
     myFile.print(",");
     myFile.print(heading);
     myFile.print(","); 
     myFile.print(x1,9);
     myFile.print(","); 
     myFile.println(y1,9);    

    myFile.print(lon6,9);
    myFile.print(",");
    myFile.print(lat6,9);
    myFile.print(",");
    myFile.print(sum6,1);
     myFile.print(",");
     myFile.print(heading);
     myFile.print(","); 
     myFile.print(x2,9);
     myFile.print(","); 
     myFile.println(y2,9);   

    myFile.print(lon7,9);
    myFile.print(",");
    myFile.print(lat7,9);
    myFile.print(",");
    myFile.print(sum7,1);
     myFile.print(",");
     myFile.print(heading);
     myFile.print(","); 
     myFile.print(x3,9);
     myFile.print(","); 
     myFile.println(y3,9);      
    
    myFile.close();
    
}

void gpsCalculate() {    //calculate the coordinates of sensors 1-3 and 5-7 using the compass direction
  // sensor4 = center

  float heading_radX = (360-heading) * DEG_TO_RAD;  // convert heading to rad
  float heading_radY = (360-heading) * DEG_TO_RAD;
  float lat_rad = lat * DEG_TO_RAD;  // convert lat to rad
  float conversionFactorX = (1/(cos(lat_rad)*111319.491666666));  // 40075017m /360°
  //float conversionFactorY = (1/111195.055555555);
  float conversionFactorY = 8.993205633*10E-6;

  x3 = ((cos(heading_radX) * distance * 3.0)*conversionFactorX);
  x2 = ((cos(heading_radX) * distance * 2.0)*conversionFactorX);
  x1 = ((cos(heading_radX) * distance)*conversionFactorX);
  y3 = ((sin(heading_radY) * distance * 3.0)*conversionFactorY);
  y2 = ((sin(heading_radY) * distance * 2.0)*conversionFactorY);
  y1 = ((sin(heading_radY) * distance)*conversionFactorY);

    lon1 = lon - x3; lat1 = lat + y3;  // add/remove values to/from measured GPS coordinates
    lon2 = lon - x2; lat2 = lat + y2;
    lon3 = lon - x1; lat3 = lat + y1;
    lon4 = lon; lat4 = lat;  // (center sensor)
    lon5 = lon + x1; lat5 = lat - y1;
    lon6 = lon + x2; lat6 = lat - y2;
    lon7 = lon + x3; lat7 = lat - y3;
}

// funktioniert ?
/* 360-heading x und y
 */
